/*
 * 文 件 名:  DynamicTask
 * 版    权:  浩鲸云计算科技股份有限公司
 * 描    述:  <描述>
 * 修 改 人:  0027008907/lv.wenbin
 * 修改时间:  2020/7/22
 * 跟踪单号:  <跟踪单号>
 * 修改单号:  <修改单号>
 * 修改内容:  <修改内容>
 */
package com.huashe.citybrain.commons.base.task;

import com.huashe.citybrain.commons.base.exception.Assert;
import com.huashe.citybrain.commons.base.exception.BusinessException;
import org.apache.commons.lang3.StringUtils;
import org.quartz.CronExpression;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.context.annotation.Configuration;
import org.springframework.scheduling.annotation.SchedulingConfigurer;
import org.springframework.scheduling.config.CronTask;
import org.springframework.scheduling.config.ScheduledTaskRegistrar;

import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ScheduledFuture;

/**
 * 动态定时器管理
 * <功能详细描述>
 *
 * @author 0027008907/lv.wenbin
 * @version [版本号, 2020/7/22]
 * @see [相关类/方法]
 * @since [产品/模块版本]
 */
@Configuration
public class DynamicTask implements SchedulingConfigurer {
    /**
     * 日志
     */
    private static final Logger log = LoggerFactory.getLogger(DynamicTask.class);

    private volatile ScheduledTaskRegistrar registrar;

    private final ConcurrentHashMap<String, ScheduledFuture<?>> scheduledFutures = new ConcurrentHashMap<>();

    private final ConcurrentHashMap<String, CronTask> cronTasks = new ConcurrentHashMap<>();

    /**
     * 添加定时任务
     *
     * @param timeTask 定时任务
     */
    public void add(TimingTask timeTask) {
        try {
            Assert.notNull(timeTask, -1, "任务对象为空");
            Assert.notEmpty(timeTask.getExpression(), -1, "定时器 cron 表达式为空");
            Assert.isTrue(CronExpression.isValidExpression(timeTask.getExpression()), -1,
                "定时任务DynamicTask cron表达式不合法:" + timeTask.getExpression());

            // 任务已存在
            if (scheduledFutures.containsKey(timeTask.getTaskId())) {
                CronTask old = cronTasks.get(timeTask.getTaskId());

                // 如果配置一致，则不需要重新创建定时任务
                if (StringUtils.equals(old.getExpression(), timeTask.getExpression())) {
                    log.warn("定时任务 taskId:{}, cron:{} 已经存在,不重复添加.", timeTask.getTaskId(), timeTask.getExpression());
                    return;
                }
                // 如果策略执行时间发生了变化，则取消当前策略的任务
                else {
                    log.info("定时任务 taskId:{}, cron:{} 表达式更新为 corn:{},重置任务.", timeTask.getTaskId(), old.getExpression(),
                        timeTask.getExpression());
                    scheduledFutures.remove(timeTask.getTaskId()).cancel(false);
                    cronTasks.remove(timeTask.getTaskId());
                }
            }

            CronTask task = new CronTask(timeTask, timeTask.getExpression());
            ScheduledFuture<?> future = registrar.getScheduler().schedule(task.getRunnable(), task.getTrigger());

            cronTasks.put(timeTask.getTaskId(), task);
            scheduledFutures.put(timeTask.getTaskId(), future);
            log.info("定时任务 taskId:{}, cron:{} 添加成功.", timeTask.getTaskId(), timeTask.getExpression());
        }
        catch (Exception e) {
            log.error("add TimingTask fail!", e);
            throw new BusinessException(-1, e.getMessage(), e);
        }
    }

    @Override
    public void configureTasks(ScheduledTaskRegistrar registrar) {
        this.registrar = registrar;
    }
}